#include <os2.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "filefilt.h"

class _Export DIFMesaFileFilter : public MesaFileFilter {
public:
    DIFMesaFileFilter() {};
    virtual ~DIFMesaFileFilter() {};

    virtual void *importFile(void *data ,int len,int &outlen,void *model,const char *f);
    virtual void *exportFile(void *data ,int len,int &outlen,void *model,const char *f);

    virtual const char *getName() {return "DIF File";};
    virtual const char *getExtension() {return "DIF";};

    // This filter can do both import and export
    virtual int canImport() { return 1;};
    virtual int canExport() { return 1;};

    // we'll base the export filter on the WK1 filter.  Several reasons for this:
    // 1) The WK1 format is well documented on the net
    // 2) Mesa spits WK1 cells out in forward order instead of
    //    backwards like it does M2 files.  Makes it easier to do the
    //    export
    // 3) It may actually be faster as Mesa will only consider the 
    //    first layer when creating the WK1 file.
    // 4) DIF files don't have formatting info so we don't really
    //    care about the advanced formatting anyway.
    virtual int getExportDataType() { return 1;};
};

extern "C" {
    MesaFileFilter *getMesaFileFilter(int num);
}



DIFMesaFileFilter theFilter;
// this is the prcoedure that Mesa will call to get a list of the file 
// filters supported by this DLL.
MesaFileFilter * _Export getMesaFileFilter(int num)
{
    if (num == 0) {
        return &theFilter;
    } // endif
    return NULL;
}


// bunch of utilities to help making the returned stream easier
static char *writeInt(char *data, int &len, int &max, int num)
{
    if ((len+4) >= max) {
        max *= 2;
        data = (char*)MRealloc(data,max);
    } // endif
    ((int *)(&data[len]))[0] = num;
    len += sizeof(int);
    return data;
}
static char *writeDouble(char *data, int &len, int &max, double num)
{
    if ((len+4) >= max) {
        max *= 2;
        data = (char*)MRealloc(data,max);
    } // endif
    ((double *)(&data[len]))[0] = num;
    len += sizeof(double);
    return data;
}
static char *writeChar(char *data, int &len, int &max, char num)
{
    if (len+1 >= max) {
        max *= 2;
        data = (char*)MRealloc(data,max);
    } // endif
    data[len] = num;
    len++;
    return data;
}
static char *writeShort(char *data, int &len, int &max, short num)
{
    if ((len+2) >= max) {
        max *= 2;
        data = (char*)MRealloc(data,max);
    } // endif
    ((short *)(&data[len]))[0] = num;
    len += sizeof(short);
    return data;
}
static char *write(char *data, int &len, int &max, const char *d, int dlen)
{
    while ((len+dlen) >= max) {
        max *= 2;
        data = (char*)MRealloc(data,max);
    } // endif
    memcpy(&data[len],d,dlen);
    len += dlen;
    return data;
}

// helper routines for skipping over areas we don't care about
inline const char *passWS(const char *s)
{
    while (*s == ' ' || *s == '\r' || *s == '\n') {
        s++;
    } // endwhile
    return s;
}
inline const char *nextLine(const char *s)
{
    while (*s && *s != '\r' && *s != '\n') {
        s++;
    } // endwhile
    return passWS(s);
}


void *DIFMesaFileFilter :: importFile(void *dataIn ,int len,int &outlen,void *model,const char *f) 
{
    outlen = 0;
    const char *indata = (const char *)dataIn;
    indata = passWS(indata);
    if (strncmp(indata,"TABLE",5)) return NULL;

    int max = 1024;
    char *data = (char *)MMalloc(max);

    // write the header
    data = write(data,outlen,max,"Mesa JDK",8);

    // the BOF record
    data = writeChar(data,outlen,max,1);    //BOF
    data = writeShort(data,outlen,max,14);  //12 bytes in size
    data = writeInt(data,outlen,max,3);     // version 3
    data = writeInt(data,outlen,max,341);   // build 341 (2.20)
    data = writeInt(data,outlen,max,850);   // codepage 850
    
    int doneHeader = 0;
    char *tablename = NULL;
    int numCols = 0;
    int numRows = 0;
    do {
        if (!strncmp(indata,"TABLE",5)) {
            indata = nextLine(indata);
            indata = nextLine(indata);
            if (*indata == '"') {
                indata++;
            } // endif
            const char *tmp = indata;
            while (*tmp != '"' && *tmp != '\n' && *tmp != '\r') {
                tmp++;
            } // endwhile
            tablename = (char *)MMalloc(tmp-indata+2);
            strncpy(tablename,indata,tmp-indata);
            if (strlen(tablename) > 15) tablename[15] = 0;
            indata = nextLine(indata);
        } else if (!strncmp(indata,"VECTORS",7)) {
            indata = nextLine(indata);
            if (*indata == '0' && indata[1] == ',') {
                indata += 2;
                numCols = atoi(indata);
            } // endif
            indata = nextLine(indata);
            indata = nextLine(indata);
        } else if (!strncmp(indata,"TUPLES",6)) {
            indata = nextLine(indata);
            if (*indata == '0' && indata[1] == ',') {
                indata += 2;
                numRows = atoi(indata);
            } // endif
            indata = nextLine(indata);
            indata = nextLine(indata);
        } else if (!strncmp(indata,"DATA",4)) {
            doneHeader = 1;
            indata = nextLine(indata);
            indata = nextLine(indata);
            indata = nextLine(indata);
        } else {
            // Unknown/unsupported header token
            // should be 3 lines long.  Skip It.
            indata = nextLine(indata);
            indata = nextLine(indata);
            indata = nextLine(indata);
        } // endif
    } while ( !doneHeader );


    /*
     Write out the layer token.  We can use the layer token to 
     specify the number of rows and columns as well as use the table name
     as the layer name
     */
    data = writeChar(data,outlen,max,12);    //LAYER
    data = writeShort(data,outlen,max,52+(tablename?strlen(tablename)+1:2));  
                                            //54+ bytes in size
    data = writeChar(data,outlen,max,0);    //reserved
    data = writeInt(data,outlen,max,0xFFFFFF);//bkgcolor
    data = writeInt(data,outlen,max,0xcccccc);//rulercolor
    data = writeInt(data,outlen,max,0xcccccc);//gridcolor
    data = writeShort(data,outlen,max,1);//rulerfont
    data = writeChar(data,outlen,max,1);    //grid
    if (tablename) {
        data = writeShort(data,outlen,max,strlen(tablename)+1);//name
        data = write(data,outlen,max,tablename,strlen(tablename)+1);
    } else {
        data = writeInt(data,outlen,max,0xFFFFFFFF);//name
    } // endif
    data = writeChar(data,outlen,max,0);    //protection
    data = writeInt(data,outlen,max,numRows+1);//nrows
    data = writeShort(data,outlen,max,numCols+1);//ncols
    data = writeChar(data,outlen,max,0);    //reserved
    data = writeShort(data,outlen,max,0);//comment
    data = writeChar(data,outlen,max,0);    //printinfo
    data = writeChar(data,outlen,max,0);    //image
    data = writeShort(data,outlen,max,8);//col ruler
    data = writeShort(data,outlen,max,72);
    data = writeInt(data,outlen,max,0xFFFFFFFF);
    data = writeShort(data,outlen,max,8);//row ruler
    data = writeShort(data,outlen,max,18);
    data = writeInt(data,outlen,max,0xFFFFFFFF);
    data = writeInt(data,outlen,max,0);//num titles

    /*
     Since we used 1 for the ruler font index, we need to 
     write out at least 2 font tokens.  Font 0 will be the default
     font and font 1 will be used for the rulers.
    */    
    data = writeChar(data,outlen,max,7);    //FONT
    data = writeShort(data,outlen,max,18);  //18 bytes in size
    data = writeInt(data,outlen,max,12);
    data = writeShort(data,outlen,max,10);
    data = write(data,outlen,max,"Helvetica",10);

    data = writeChar(data,outlen,max,7);    //FONT
    data = writeShort(data,outlen,max,23);  //23 bytes in size
    data = writeInt(data,outlen,max,14);
    data = writeShort(data,outlen,max,15);
    data = write(data,outlen,max,"Helvetica Bold",15);

    /*
     Go through the data section of the file and create cell tokens
    */
    int doneData = 0;
    int curRow = 0;
    int curCol = 0;
    int hasData = 0;
    do {
        switch (*indata) {
        case '-':
            // -1: control token/special data value
            indata = nextLine(indata);
            if (!strncmp(indata,"BOT",3)) {
                if (hasData) curRow++;
                hasData = 1;
                curCol = 0;
            } else if (!strncmp(indata,"EOD",3)) {
                doneData = 1;
            } else {
                //error
            } // endif
            indata = nextLine(indata);
            break;
        case '0':
            {
                // 0: numeric data
                hasData = 1;
                indata += 2;
                double dbl = atof(indata);
                indata = nextLine(indata);
                int error = -1;
                if (*indata == 'N') {
                    // NA condition
                    error = 0;
                } else if (*indata == 'E') {
                    // error condition
                    error = 30; // general error number in Mesa
                } // endif

                // CELL record
                data = writeChar(data,outlen,max,3);    //CELL
                if (error == -1) {
                    // cell with valid number in it
                    data = writeShort(data,outlen,max,23);  //size of 23
                    data = writeChar(data,outlen,max,0);    //address flag
                    data = writeShort(data,outlen,max,0);   //layer
                    data = writeShort(data,outlen,max,curCol); //column
                    data = writeInt(data,outlen,max,curRow);//row number
                    data = writeShort(data,outlen,max,2 | (3 << 7)); //flags
                    data = writeShort(data,outlen,max,0);   //format
                    data = writeDouble(data,outlen,max,dbl);
                } else {
                    // either a blank cell or a cell with an error
                    data = writeShort(data,outlen,max,17);  //size of 17
                    data = writeChar(data,outlen,max,0);    //address flag
                    data = writeShort(data,outlen,max,0);   //layer
                    data = writeShort(data,outlen,max,curCol); //column
                    data = writeInt(data,outlen,max,curRow);//row number
                    data = writeShort(data,outlen,max,0);//flags
                    data = writeShort(data,outlen,max,0);   //format
                    data = writeShort(data,outlen,max,error);//error number
                } // endif
                curCol++;
                indata = nextLine(indata);
            }
            break;
        case '1':
            {
                //1: string data
                hasData = 1;
                indata = nextLine(indata);
                indata++;
                const char *tmp = indata;
                while (*tmp != '"') tmp++;

                //if the cell is empty, don't bother writing it out
                if (tmp != indata) {
                    // CELL record
                    data = writeChar(data,outlen,max,3);    //CELL
                    data = writeShort(data,outlen,max,17+tmp-indata+1);  //size of 23
                    data = writeChar(data,outlen,max,0);    //address flag
                    data = writeShort(data,outlen,max,0);   //layer
                    data = writeShort(data,outlen,max,curCol); //column
                    data = writeInt(data,outlen,max,curRow);//row number
                    data = writeShort(data,outlen,max,1); //flags
                    data = writeShort(data,outlen,max,0);   //format
                    data = writeShort(data,outlen,max,tmp-indata+1);//string len
                    data = write(data,outlen,max,indata,tmp-indata);
                    data = writeChar(data,outlen,max,0);    //NULL terminate
                } // endif
                curCol++;
                indata = nextLine(indata);
            }
            break;
        default:
            // error
            curCol++;
            indata = nextLine(indata);
            indata = nextLine(indata);
            break;
        } // endswitch

    } while ( !doneData && ((indata - (const char *)dataIn)< len));
    MFree(tablename);

    // EOF record
    data = writeChar(data,outlen,max,2);    //EOF
    data = writeShort(data,outlen,max,2);   //size of 2

    return (void *)data;
}




static int readShort(const char *d,int &len)
{
    int i = *((short*)&d[len]);
    len += sizeof(short);
    return i;
}
static double readDouble(const char *d,int &len)
{
    double i = *((double*)&d[len]);
    len += sizeof(double);
    return i;
}

void *DIFMesaFileFilter :: exportFile(void *dataIn ,int len,int &outlen,void *model,const char *f) 
{
    outlen = 0;
    int max = 2048;
    char *data = (char *)MMalloc(max);
    const char *indata = (const char *)dataIn;

    int maxRow = 0;
    int maxCol = 0;

    int curPos = 0;

    //spin through the data once to determine limits
    int token = readShort(indata,curPos);
    int nextPos = readShort(indata,curPos);
    nextPos += curPos;
    while (token != 1) {  //EOF token is 1
        switch (token) {
        case 14: //float
        case 15: //string
        case 16: //formula
            {
                curPos++;  // skip over the format
                int col = readShort(indata,curPos);
                int row = readShort(indata,curPos);
                if (col > maxCol) maxCol = col;
                if (row > maxRow) maxRow = row;
            }
            break;
        } // endswitch
        curPos = nextPos;
        token = readShort(indata,curPos);
        nextPos = readShort(indata,curPos);
        nextPos += curPos;
    } // endwhile

    // Write the header part
    strcpy(data,"TABLE\n0,1\n\"Sheet1\"\nVECTORS\n0,");
    outlen = strlen(data);
    sprintf(&data[outlen],"%d",maxCol+1);
    outlen += strlen(&data[outlen]);
    sprintf(&data[outlen],"\n\"\"\nTUPLES\n0,%d\n\"\"\nDATA\n0,0\n\"\"\n",maxRow+1);
    outlen += strlen(&data[outlen]);
    sprintf(&data[outlen],"-1,0\nBOT\n");
    outlen += strlen(&data[outlen]);

    int curRow = 0;
    int curCol = 0;
    //spin trough again outputing the data
    curPos = 0;
    token = readShort(indata,curPos);
    nextPos = readShort(indata,curPos);
    nextPos += curPos;
    while (token != 1) {  //EOF token is 1
        if ((outlen+256) > max) {
            max *= 2;
            data = (char *)MRealloc(data,max);
        } // endif
        switch (token) {
        case 14: //float
        case 15: //string
        case 16: //formula
            {
                curPos++;  // skip over the format
                int col = readShort(indata,curPos);
                int row = readShort(indata,curPos);
                if (col > maxCol) maxCol = col;
                if (row > maxRow) maxRow = row;
                while (row != curRow) {
                    if ((outlen+256) > max) {
                        max *= 2;
                        data = (char *)MRealloc(data,max);
                    } // endif
                    curCol = 0;
                    curRow++;
                    sprintf(&data[outlen],"-1,0\nBOT\n");
                    outlen += strlen(&data[outlen]);
                } // endif
                while (curCol != col ) {
                    if ((outlen+256) > max) {
                        max *= 2;
                        data = (char *)MRealloc(data,max);
                    } // endif
                    sprintf(&data[outlen],"1,0\n\"\"\n");
                    outlen += strlen(&data[outlen]);
                    curCol++;
                } // endwhile

                switch (token) {
                case 16: //formula
                case 14: //float
                    {
                        double d = readDouble(indata,curPos);
                        sprintf(&data[outlen],"0,%lf\nV\n",d);
                        outlen += strlen(&data[outlen]);
                    }
                    break;
                case 15: //string
                    curPos++; //skip alignment byte
                    while ((outlen+strlen(&indata[curPos])+10) > max) {
                        max *= 2;
                        data = (char *)MRealloc(data,max);
                    } // endif
                    sprintf(&data[outlen],"1,0\n\"%s\"\n",&indata[curPos]);
                    outlen += strlen(&data[outlen]);
                    break;
                };
                curCol = col+1;
            }
            break;
        } // endswitch
        curPos = nextPos;
        token = readShort(indata,curPos);
        nextPos = readShort(indata,curPos);
        nextPos += curPos;
    } // endwhile

    // end of data token
    sprintf(&data[outlen],"-1,0\nEOD\n");
    outlen += strlen(&data[outlen]);


    return (void *)data;
}

